package no.tornado.databinding;

import no.tornado.databinding.validator.ValidationStrategy;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.util.ArrayList;
import java.util.List;

public class BindingGroup {
    private List<Binding> bindings = new ArrayList<Binding>();
    private BindingStrategy defaultBindingStrategy = BindingStrategy.ONFLUSH;
    private ValidationStrategy defaultValidationStrategy = ValidationStrategy.ONCHANGE;
    private Boolean usePropertyChangeSupport = false;
    private Object defaultModel;
    private List<ChangeListener> bindingStateListeners = new ArrayList<ChangeListener>();

    public BindingGroup() {
    }

    public BindingGroup(Object defaultModel) {
        this.defaultModel = defaultModel;
    }

    public BindingGroup bind() {
        for (Binding binding : bindings)
            binding.bind();
        
        return this;
    }

    public BindingGroup unbind() {
        for (Binding binding : bindings)
            binding.unbind();

        return this;
    }

    public void flushUIToModel() {
        for (Binding binding : bindings)
            binding.flushUIToModel();
    }

    public void flushModelToUI() {
        for (Binding binding : bindings)
            binding.flushModelToUI();
    }

    public Binding add(JComponent ui, Object model, String propertyExpression) {
        Binding binding = Binding.create(ui, model, propertyExpression).bindingStrategy(defaultBindingStrategy).validationStrategy(defaultValidationStrategy);
        add(binding);
        binding.setBindingGroup(this);
        return binding;
    }

    public Binding add(JComponent ui, String propertyExpression) {
        if (defaultModel == null)
            throw new RuntimeException("Set defaultModel before creating a binding without an explicit model");
        
        Binding binding = Binding.create(ui, defaultModel, propertyExpression).bindingStrategy(defaultBindingStrategy).validationStrategy(defaultValidationStrategy);
        add(binding);
        binding.setBindingGroup(this);
        if (getUsePropertyChangeSupport())
            binding.setUsePropertyChangeSupport(true);
        return binding;
    }

    public BindingGroup add(Binding binding) {
        bindings.add(binding);

        if (binding.getModel() == null)
            binding.setModel(defaultModel);

        if (getUsePropertyChangeSupport())
            binding.setUsePropertyChangeSupport(true);

        binding.setBindingGroup(this);

        return this;
    }

    public List<Binding> getBindings() {
        return bindings;
    }

    public void setBindings(List<Binding> bindings) {
        this.bindings = bindings;
    }

    public BindingStrategy getDefaultBindingStrategy() {
        return defaultBindingStrategy;
    }

    public void setDefaultBindingStrategy(BindingStrategy defaultBindingStrategy) {
        this.defaultBindingStrategy = defaultBindingStrategy;
        for (Binding binding : bindings)
            binding.setBindingStrategy(defaultBindingStrategy);
    }

    public BindingGroup defaultBindingStrategy(BindingStrategy defaultBindingStrategy) {
        this.defaultBindingStrategy = defaultBindingStrategy;
        return this;
    }

    public ValidationStrategy getDefaultValidationStrategy() {
        return defaultValidationStrategy;
    }

    public void setDefaultValidationStrategy(ValidationStrategy defaultValidationStrategy) {
        this.defaultValidationStrategy = defaultValidationStrategy;
    }

    public BindingGroup defaultValidationStrategy(ValidationStrategy defaultValidationStrategy) {
        this.defaultValidationStrategy = defaultValidationStrategy;
        return this;
    }

    public Object getDefaultModel() {
        return defaultModel;
    }

    /**
     * Swap modelObject in all bindings that use oldModel
     * @param oldModel The model that will be replaced by newModel
     * @param newModel  The model that will replace oldModel
     */
    public void changeModelObjectForBindings(Object oldModel, Object newModel) {
        for (Binding binding : bindings)
            if (oldModel.equals(binding.getModel()))
                binding.setModel(newModel);
    }

    /**
     * Set a defaultModel to use for bindings. Old bindings that use the old defaultModel will
     * automatically get the new defaultModel instead.
     * @param defaultModel The new defaultModel
     */
    public void setDefaultModel(Object defaultModel) {
        // Switch out this modelobject if used in bindings as well
        if (this.defaultModel != null)
            changeModelObjectForBindings(this.defaultModel, defaultModel);

        this.defaultModel = defaultModel;
    }

    public BindingGroup defaultModel(Object defaultModel) {
        setDefaultModel(defaultModel);
        return this;
    }
    
    public void bindingInvalid(Binding binding) {
        ChangeEvent evt = new ChangeEvent(binding);
        for (ChangeListener listener : bindingStateListeners)
            listener.stateChanged(evt);
    }

    public void bindingValid(Binding binding) {
        ChangeEvent evt = new ChangeEvent(binding);
        for (ChangeListener listener : bindingStateListeners)
            listener.stateChanged(evt);
    }

    public boolean isValid() {
        for (Binding binding : bindings)
            if (!binding.isValid())
                return false;

        return true;
    }

    public void addBindingStateListener(ChangeListener listener) {
        bindingStateListeners.add(listener);
    }

    public void removeBindingStateListener(ChangeListener listener) {
        bindingStateListeners.remove(listener);
    }

    public Boolean getUsePropertyChangeSupport() {
        return usePropertyChangeSupport;
    }

    public void setUsePropertyChangeSupport(Boolean usePropertyChangeSupport) {
        this.usePropertyChangeSupport = usePropertyChangeSupport;
        for (Binding binding : bindings)
            binding.setUsePropertyChangeSupport(usePropertyChangeSupport);
    }

    public BindingGroup usePropertyChangeSupport(Boolean usePropertyChangeSupport) {
        setUsePropertyChangeSupport(usePropertyChangeSupport);
        return this;
    }
}
